/** 
 *   Test: cmd_tools
 *
 *   (c) Copyright 2007 Federico Bareilles <fede@iar.unlp.edu.ar>,
 *   Instituto Argentino de Radio Astronomia (IAR).
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *     
 *   The author does NOT admit liability nor provide warranty for any
 *   of this software. This material is provided "AS-IS" in the hope
 *   that it may be useful for others.
 *
 *   29/03/2007: Created.
 **/

#define _GNU_SOURCE


#include <stdio.h>
#include <stdint.h>
#include <limits.h>
#include <unistd.h>
#include <stdlib.h>
#include <getopt.h>
#include <time.h>

#include "my_stdio.h"
#include "cmd_tools.h"
#include "iar_bus.h"
#include "ib_dev.h"
#include "socket_client.h"

#define TEST_IP "127.0.0.1"
#define TEST_PORT 1100

#define CH_MAGICID 0x0010
#define CH_MAGIC_WHO  (CH_MAGICID | 0x1)
#define CH_MAGIC_INFO (CH_MAGICID | 0x2)

typedef struct _message_head {
	uint32_t magic_id:16;
	char name[10];
	uint32_t time;	
} __attribute__((__may_alias__)) MESSAGE_HEAD;


#define MESSAGE_MAGIC_ID(m) (m)->magic_id
#define MESSAGE_NAME(m) (m)->name
#define MESSAGE_TIME(m) (m)->time
#define MESSAGE_IF(m) (MESSAGE_MAGIC_ID(m) == CH_MAGICID ? 1 : 0)
#define MSSAGE_WHO_IF(m) (MESSAGE_MAGIC_ID(m) == CH_MAGIC_WHO ? 1 : 0)
#define MSSAGE_INFO_IF(m) (MESSAGE_MAGIC_ID(m) == CH_MAGIC_INFO ? 1 : 0)
#define MESSAGE_SET_ID(m) MESSAGE_MAGIC_ID(m) = CH_MAGICID
#define MESSAGE_SET_WHO(m) MESSAGE_MAGIC_ID(m) = CH_MAGIC_WHO
#define MESSAGE_SET_INFO(m) MESSAGE_MAGIC_ID(m) = CH_MAGIC_INFO

#define PK_MESSAGE_GET(d) (IB_PK_DATA(d) + sizeof(MESSAGE_HEAD))
#define PK_MESSAGE_SIZE(d) (IB_PK_SIZE(d) - sizeof(MESSAGE_HEAD))
#define PK_GET_CMD(d) IB_PK_CMD(d)

//#define PK_MESSAGE_SIZE_IF_VALID(d) (IB_PK_SIZE(d) < IB_PACKET_DATA_LEN?1:0)


IBFD *fd_sock = NULL;
uint32_t my_addr = 0;
uint32_t addr_to = 0;

int get_data(int fd, void* data);

int com_msa    __P((char *));
int com_to __P((char *));
int com_addr __P((char *));
int com_who __P((char *));
int com_beep __P((char *));
int com_core_shutdown __P((char *));
int com_devs_list __P((char *));
int com_set_mcast_list __P((char *));

COMMAND my_commands[] = {
  { "cd",       0, 0, com_chdir, "[path]", "Change working directory" },
  { "export",   0, 0, com_export, "name=value", 
    "Change or add an environment variable" },	
  { "help",     0, 0, com_help, "[cmd]", "Display this text" },
  { "?",        0, 1, com_help, "[cmd]", "Synonym for `help'" },
  { "quit",     0, 0, com_quit, "", "Quit ..." },
  { "q",        0, 1, com_quit, "", "Quit :-)" },
  { "msa",      1, 1, com_msa, "[message]", "Send message to all or address" },
  { "toaddr",   0, 1, com_to, "[address]", 
    "Get/Set to address value (0 to all)" },
  { "addr",     0, 1, com_addr, "", "Show my address" },
  { "who",      0, 1, com_who, "", "List connected users" },
  { "beep",     1, 1, com_beep, "", "Send \"beep\" to all or address" },
  { "coredown", 1, 1, com_core_shutdown,"", "Shut down IAR Bus core" },
  { "devs_list", 1, 1, com_devs_list, "", "Show curren devices in IAR Bus" },
  { "mcast_list", 1, 1, com_set_mcast_list, "[addr list]", 
    "Setting/Removing multi cast list in Bus"},    
  { (char *)NULL, 0, 0, (rl_icpfunc_t *)NULL, (char *) NULL, (char *)NULL }
};




const char *help_text[] ={
#       ifdef VERBOSE_MODE
        "  -v\t\t - Verbose mode (incremental).",
#       endif
        "  -s {server}\t - Set server name or IP addrs (default: "TEST_IP")",
//        "  -n [--interac] - Not interactive mode.",
//        "  -r\t\t - Don't use \"read line\" history.",
	"  --link\t - Make symbolic links for each command.",
	"  --unlink\t - Remove symbolic links for commands.",
        "  -h\t\t - This help !!!.",
        ""
};


int help(void)
{
	int i=0;

        while((char) *help_text[i] != 0) fprintf(stderr,"%s\n",help_text[i++]);
        fprintf(stderr,"\n");

	return 0;
}


int send_mess_type_to(IBFD *dc, int type, int addr, char *msg)
{
	int size = sizeof(MESSAGE_HEAD) + (msg != NULL?strlen(msg):0); 
	char *data = (char *) malloc(size);
	MESSAGE_HEAD *m = (MESSAGE_HEAD *) data;
	char *user = getenv("USER");
	
	memset(data, 0, size);
	MESSAGE_TIME(m) = time(NULL);
	MESSAGE_MAGIC_ID(m) = type;
	if (user != NULL) {
		memcpy(MESSAGE_NAME(m), user,
		       strlen(user) > 10 ? 10 : strlen(user));
	}
	if ( msg != NULL) 
		memcpy(data+sizeof(MESSAGE_HEAD), msg, 
		       size - sizeof(MESSAGE_HEAD));
	if (addr) {
		if (ibus_send_data_to(dc, addr, data, size) != 0) {
			printf("Send message Fail.\n");
		}
	} else {
		ibus_send_data( dc, data, size);
	}
	free(data);
	
	return 0;
}


int declare(IBFD *dc, int addr)
{
	char hostaname[HOST_NAME_MAX];
	gethostname(hostaname, HOST_NAME_MAX);

	return send_mess_type_to(dc, CH_MAGIC_INFO, addr, hostaname);
}


int main(int argc, char * const argv[])
{
	char *server_ip = TEST_IP;
	int ret;
	int op;
	struct option const longopts[] =
        {
                {"help", 0, 0, 'h'},
                {"verbose", 0, 0, 'v'}, 
		{"link", 0, 0, 201},
		{"unlink", 0, 0, 202},
                {0,  0,  0,  0}
        };
	SET_VERBOSE_FLAG(PERR);
	cmd_tools_set_command(my_commands);
	while  ((op = getopt_long(argc, argv, "hvs:",longopts, 0)) != EOF)
		switch( op ){
		case 's':
			server_ip = optarg;
			break;
		case 'v':
			INCREMENT_VERBOSE_FLAG();
			break;
		case 'V':
                        //version(program_name);
                        exit(-1);
                        break;
		case 201:
			cmd_tools_link_commands_to(argv[0] /*program_name*/ );
			exit(0); /* ok */ 
		case 202:
			cmd_tools_un_link_commands();
			exit(0); /* ok */
		case 'h':
		default:
			printf("Use %s [-v (incr.)]\n", argv[0]);
			help();
			exit(-1);
			break;
		}

	fd_sock = ibus_client_connect(server_ip, TEST_PORT);
	if(fd_sock == NULL) {
		PVERB(PERR,"Can't connect to server (%s).", server_ip);
		return -1;
	}
	ibus_set_msg_confirmation(fd_sock, 1);
#if 1
	if ( ibus_bcast_on( fd_sock ) != 0 ) {
		PVERB(PERR,"Can't start broadcast mode.");
	}
#endif
	ibus_get_addr(fd_sock, &my_addr);
	printf("My address is: 0x%03x\n", my_addr);

	cmd_tools_set_prompt("chat> ");
	cmd_tools_use_history("cmd");
	cmd_tools_use_baner(1);
	cmd_tools_add_fd_callback(IBFD_GET(fd_sock), get_data, fd_sock);
	declare(fd_sock, 0);

	ret = cmd_tools_main_loop(argc, argv);

	ibus_remove(fd_sock);
	ibus_client_close(fd_sock);

	return ret;
}


int com_msa(char *cmd)
{
	return send_mess_type_to(fd_sock, CH_MAGICID, addr_to, cmd);
}


int com_beep(char *cmd)
{
	char beep[] = {0x07, 0};
	return send_mess_type_to(fd_sock, CH_MAGICID, addr_to, beep);
}


int com_who(char *cmd)
{
	return send_mess_type_to(fd_sock, CH_MAGIC_WHO, 0, NULL);
}


int com_to(char *cmd)
{
	if ( *(cmd) != 0 ) 
		addr_to = (uint32_t) strtol(cmd, NULL, 16); 
	else
		printf("0x%04x\n", addr_to);

	return 0;
}


int com_addr(char *cmd)
{
	printf("0x%04x\n", my_addr);

	return 0;
}


int get_data(int fd, void* data)
{
	int ret = 1;
	IB_PK d;
	IBFD *dc = (IBFD *) data;
	MESSAGE_HEAD *m = (MESSAGE_HEAD *) IB_PK_DATA(&d);
	struct tm *t;
	time_t mtime;
	char name[11];

	memset(m, 0, sizeof(MESSAGE_HEAD));
	if ((ret = ibus_read(dc, &d)) > 0) {
		if ( ! IB_PK_SIZE_IF_VALID(&d)) {
			return -1;
		}
		if (MESSAGE_IF(m)) {
			memcpy(name, MESSAGE_NAME(m), 10);
			name[10] = 0;
			mtime = MESSAGE_TIME(m);
			t = localtime(&mtime);
			*(PK_MESSAGE_GET(&d)+PK_MESSAGE_SIZE(&d)) = 0;
			printf("\r[%02d:%02d:%02d] %s (0x%03x): %s\n",
			       t->tm_hour,t->tm_min,t->tm_sec, 
			       name, IB_PK_ADDR(&d),
			       PK_MESSAGE_GET(&d) );
		} else if (MSSAGE_WHO_IF(m)) {
			declare(dc, IB_PK_ADDR(&d));
		} else if (MSSAGE_INFO_IF(m)) {
			memcpy(name, MESSAGE_NAME(m), 10);
			name[10] = 0;
			*(PK_MESSAGE_GET(&d)+PK_MESSAGE_SIZE(&d)) = 0;
			printf("\r%-10s 0x%03x (%s)", name, IB_PK_ADDR(&d),
				PK_MESSAGE_GET(&d) );
			if ( IB_PK_ADDR(&d) == addr_to) 
				printf(" *");
			printf("\n");
		} else if ( PK_GET_CMD(&d) == IB_CM_SHUTDOWN ) {
			printf("\rIAR Bus is shut down.  \n");
			com_quit(NULL);	
		} else {
			fprintf(stderr, "\rPK unknow: ");
			f_ib_pk_print(stderr, d);
		}
			
	}

	return ret;
}


int com_core_shutdown(char *cmd)
{
	if (ibus_shutdown(fd_sock) == 0)
		com_quit(NULL);
	else 
		printf("\rFail: Shutdown not enable in IAR Bus.  \n");	

	return 0;
}


int com_devs_list(char * cmd)
{
	int i;
	DEVS *dv = NULL;
	DEVS *dvp;
	int len = ibus_who(fd_sock, &dv);
	dvp = dv;
	if (len > 0) {
		for(i = 0; i < len; i+= sizeof(DEVS)) {
			printf("addr: 0x%x, st: 0x%x [br: %s, busy: %s, r: %s, w: %s, new: %s, alive: %s]\n",
			       DEVS_ADDR(dvp), DEVS_STATUS(dvp),
			       (DEVS_STATUS(dvp) & IB_DEV_BCAST_ON ?"on ":"off"),
			       (DEVS_STATUS(dvp) & IB_DEV_BUSY ?"on ":"off"),
			       (DEVS_STATUS(dvp) & IB_DEV_READ ?"on ":"off"),
			       (DEVS_STATUS(dvp) & IB_DEV_WRITE ?"on ":"off"),
			       (DEVS_STATUS(dvp) & IB_DEV_NEWPK ?"on ":"off"),
			       (DEVS_STATUS(dvp) & IB_DEV_ALIVE ?"on ":"off")
				);
			dvp++;
		}
		free(dv);
	}

	return len;
}
	

int com_set_mcast_list(char *cmd)
{
#if 0
	ADDR ad_fix_list[] = {
		{0x403},
		{0x404},
		{0x405},
		{IB_NULL_ADDR},
	};
#endif
	int ret;
	char *np; 
        char *endp = stripwhite(cmd);
	ADDR *ad = (ADDR *) malloc(ADDR_LIMIT * sizeof(ADDR));
	int i = 0;

	if ( *endp ) {
		do {
			np = endp;
			ADDR_GET(ad+(i++)) = (uint16_t) strtol(np, &endp, 16 );
		} while(*endp != '\0');
		ADDR_GET(ad+i) = IB_NULL_ADDR;
		ret = ibus_set_mcast_list(fd_sock, ad);
	} else {
		ret = ibus_remove_mcast_list(fd_sock);
	}

	free(ad);

	return ret;
}
